// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    int::{KnownSize, UnknownSize, UNKNOWN_SIZE},
    intern::{Intern, Interned},
    ty::{CanonicalType, StaticType, Type},
    util::GenericConstUsize,
};
use std::{marker::PhantomData, ops::Index};

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct Array<T: Type, Len: GenericConstUsize = UnknownSize> {
    element: T,
    len: usize,
    _phantom: PhantomData<Len>,
}

#[allow(non_upper_case_globals)]
pub const Array: ArrayWithoutGenerics = ArrayWithoutGenerics;

impl<T: Type, Len: GenericConstUsize> Array<T, Len> {
    pub const fn element(&self) -> &T {
        &self.element
    }
    pub const fn len(&self) -> usize {
        if Len::VALUE != UNKNOWN_SIZE {
            debug_assert!(self.len == Len::VALUE);
            Len::VALUE
        } else {
            self.len
        }
    }
}

impl<T: Type, Len: KnownSize> Array<T, Len> {
    pub const fn new_static(element: T) -> Self {
        Self {
            element,
            len: Len::VALUE,
            _phantom: PhantomData,
        }
    }
}

impl<T: StaticType, Len: KnownSize> StaticType for Array<T, Len> {
    fn static_type() -> Self {
        Self::new_static(T::static_type())
    }
}

impl<T: Type> Array<T> {
    pub const fn new_dyn(element: T, len: usize) -> Self {
        assert!(len != UNKNOWN_SIZE);
        Self {
            element,
            len,
            _phantom: PhantomData,
        }
    }
}

impl<T: Type, Len: GenericConstUsize> Type for Array<T, Len> {
    fn canonical(&self) -> CanonicalType {
        CanonicalType::Array(Intern::intern_sized(Array::new_dyn(
            self.element().canonical(),
            self.len(),
        )))
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug, Default)]
pub struct ArrayWithoutGenerics;

impl<T: Type> Index<T> for ArrayWithoutGenerics {
    type Output = ArrayWithoutLen<T>;

    fn index(&self, element: T) -> &Self::Output {
        Interned::<_>::into_inner(Intern::intern_sized(ArrayWithoutLen { element }))
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct ArrayWithoutLen<T: Type> {
    element: T,
}

impl<T: Type> Index<usize> for ArrayWithoutLen<T> {
    type Output = Array<T>;

    fn index(&self, len: usize) -> &Self::Output {
        Interned::<_>::into_inner(Intern::intern_sized(Array::new_dyn(self.element, len)))
    }
}
