// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    intern::{Intern, Interned},
    ty::{CanonicalType, StaticType, Type},
};
use hashbrown::HashSet;

#[derive(Copy, Clone, Debug, Hash, PartialEq, Eq)]
pub struct BundleField {
    pub name: Interned<str>,
    pub flipped: bool,
    pub ty: CanonicalType,
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct Bundle {
    fields: Interned<[BundleField]>,
}

impl Bundle {
    #[track_caller]
    pub fn new(fields: &[BundleField]) -> Self {
        let fields = fields.intern();
        let mut names = HashSet::new();
        for field in fields {
            assert!(
                names.insert(field.name),
                "duplicate field name: {:?}",
                field.name
            );
        }
        Self { fields }
    }
}

impl Type for Bundle {
    fn canonical(&self) -> CanonicalType {
        CanonicalType::Bundle(*self)
    }
}

macro_rules! impl_tuples {
    ([$({#[num = $num:literal] $var:ident: $T:ident})*] []) => {
        impl<$($T: Type,)*> Type for ($($T,)*) {
            fn canonical(&self) -> CanonicalType {
                let ($($var,)*) = self;
                Bundle::new(&[$(BundleField { name: stringify!($num).intern(), flipped: false, ty: $var.canonical() }),*]).canonical()
            }
        }
        impl<$($T: StaticType,)*> StaticType for ($($T,)*) {
            fn static_type() -> Self {
                $(let $var = $T::static_type();)*
                ($($var,)*)
            }
        }
    };
    ([$($lhs:tt)*] [$rhs_first:tt $($rhs:tt)*]) => {
        impl_tuples!([$($lhs)*] []);
        impl_tuples!([$($lhs)* $rhs_first] [$($rhs)*]);
    };
}

impl_tuples! {
    [] [
        {#[num = 0] a: A}
        {#[num = 1] b: B}
        {#[num = 2] c: C}
        {#[num = 3] d: D}
        {#[num = 4] e: E}
        {#[num = 5] f: F}
        {#[num = 6] g: G}
        {#[num = 7] h: H}
        {#[num = 8] i: I}
        {#[num = 9] j: J}
        {#[num = 10] k: K}
        {#[num = 11] l: L}
    ]
}
