// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    int::{Bool, IntType, SInt, SIntValue, UInt, UIntValue},
    intern::{Intern, Interned},
    ty::{CanonicalType, Type},
    util::GenericConstUsize,
};

pub mod ops;

macro_rules! expr_enum {
    (
        pub enum $ExprEnum:ident {
            $($Variant:ident($VariantTy:ty),)*
        }
    ) => {
        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub enum $ExprEnum {
            $($Variant($VariantTy),)*
        }

        $(impl From<$VariantTy> for $ExprEnum {
            fn from(v: $VariantTy) -> Self {
                Self::$Variant(v)
            }
        })*
    };
}

expr_enum! {
    pub enum ExprEnum {
        UIntLiteral(Interned<UIntValue>),
        SIntLiteral(Interned<SIntValue>),
        BoolLiteral(bool),
        NotU(ops::NotU),
        NotS(ops::NotS),
        NotB(ops::NotB),
        Neg(ops::Neg),
        BitAndU(ops::BitAndU),
        BitAndS(ops::BitAndS),
        BitAndB(ops::BitAndB),
        BitOrU(ops::BitOrU),
        BitOrS(ops::BitOrS),
        BitOrB(ops::BitOrB),
        BitXorU(ops::BitXorU),
        BitXorS(ops::BitXorS),
        BitXorB(ops::BitXorB),
        AddU(ops::AddU),
        AddS(ops::AddS),
        SubU(ops::SubU),
        SubS(ops::SubS),
        MulU(ops::MulU),
        MulS(ops::MulS),
        DivU(ops::DivU),
        DivS(ops::DivS),
        RemU(ops::RemU),
        RemS(ops::RemS),
    }
}

impl From<UIntValue> for ExprEnum {
    fn from(value: UIntValue) -> Self {
        ExprEnum::UIntLiteral(Intern::intern_sized(value))
    }
}

impl From<SIntValue> for ExprEnum {
    fn from(value: SIntValue) -> Self {
        ExprEnum::SIntLiteral(Intern::intern_sized(value))
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct Expr<T: Type> {
    __enum: Interned<ExprEnum>,
    __ty: T,
}

impl<T: Type> Expr<T> {
    pub fn expr_enum(this: Self) -> Interned<ExprEnum> {
        this.__enum
    }
    pub fn ty(this: Self) -> T {
        this.__ty
    }
    pub fn canonical(this: Self) -> Expr<CanonicalType> {
        Expr {
            __enum: this.__enum,
            __ty: this.__ty.canonical(),
        }
    }
    pub fn as_dyn_int(this: Self) -> Expr<T::Dyn>
    where
        T: IntType,
    {
        Expr {
            __enum: this.__enum,
            __ty: this.__ty.as_dyn_int(),
        }
    }
}

pub trait ToExpr {
    type Type: Type;
    fn to_expr(&self) -> Expr<Self::Type>;
}

impl<T: Type> ToExpr for Expr<T> {
    type Type = T;

    fn to_expr(&self) -> Expr<Self::Type> {
        *self
    }
}

impl<T: ?Sized + ToExpr> ToExpr for &'_ T {
    type Type = T::Type;

    fn to_expr(&self) -> Expr<Self::Type> {
        T::to_expr(self)
    }
}

impl<T: ?Sized + ToExpr> ToExpr for &'_ mut T {
    type Type = T::Type;

    fn to_expr(&self) -> Expr<Self::Type> {
        T::to_expr(self)
    }
}

impl<T: ?Sized + ToExpr> ToExpr for Box<T> {
    type Type = T::Type;

    fn to_expr(&self) -> Expr<Self::Type> {
        T::to_expr(self)
    }
}

impl<T: ?Sized + ToExpr + Send + Sync + 'static> ToExpr for Interned<T> {
    type Type = T::Type;

    fn to_expr(&self) -> Expr<Self::Type> {
        T::to_expr(self)
    }
}

impl<Width: GenericConstUsize> ToExpr for UIntValue<Width> {
    type Type = UInt<Width>;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::UIntLiteral(self.clone().as_dyn_int().intern()).intern(),
            __ty: self.ty(),
        }
    }
}

impl<Width: GenericConstUsize> ToExpr for SIntValue<Width> {
    type Type = SInt<Width>;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::SIntLiteral(self.clone().as_dyn_int().intern()).intern(),
            __ty: self.ty(),
        }
    }
}

impl ToExpr for bool {
    type Type = Bool;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::BoolLiteral(*self).intern(),
            __ty: Bool,
        }
    }
}
