// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    expr::{Expr, ExprEnum, ToExpr},
    int::{Bool, IntType, SInt, SIntValue, UInt, UIntValue, UnknownSize},
    intern::Intern,
    util::GenericConstUsize,
};
use std::ops::{Add, BitAnd, BitOr, BitXor, Div, Mul, Neg as StdNeg, Not, Rem, Sub};

macro_rules! forward_value_to_expr_unary_op_trait {
    (
        #[generics($($generics:tt)*)]
        #[value($Value:ty)]
        $Trait:ident::$method:ident
    ) => {
        impl<$($generics)*> $Trait for $Value
        where
            Expr<<$Value as ToExpr>::Type>: $Trait,
        {
            type Output = <Expr<<$Value as ToExpr>::Type> as $Trait>::Output;

            fn $method(self) -> Self::Output {
                $Trait::$method(self.to_expr())
            }
        }
    };
}

macro_rules! impl_unary_op_trait {
    (
        #[generics($($generics:tt)*)]
        fn $Trait:ident::$method:ident($arg:ident: $Arg:ty) -> $Output:ty {
            $($body:tt)*
        }
    ) => {
        impl<$($generics)*> $Trait for Expr<$Arg>
        {
            type Output = Expr<$Output>;

            fn $method(self) -> Self::Output {
                let $arg = self;
                $($body)*
            }
        }
    };
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct NotU<Width: GenericConstUsize = UnknownSize> {
    pub arg: Expr<UInt<Width>>,
}

impl<Width: GenericConstUsize> ToExpr for NotU<Width> {
    type Type = UInt<Width>;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::NotU(NotU {
                arg: Expr::as_dyn_int(self.arg),
            })
            .intern(),
            __ty: self.arg.__ty,
        }
    }
}

impl_unary_op_trait! {
    #[generics(Width: GenericConstUsize)]
    fn Not::not(arg: UInt<Width>) -> UInt<Width> {
        NotU { arg }.to_expr()
    }
}

forward_value_to_expr_unary_op_trait! {
    #[generics(Width: GenericConstUsize)]
    #[value(UIntValue<Width>)]
    Not::not
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct NotS<Width: GenericConstUsize = UnknownSize> {
    pub arg: Expr<SInt<Width>>,
}

impl<Width: GenericConstUsize> ToExpr for NotS<Width> {
    type Type = UInt<Width>;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::NotS(NotS {
                arg: Expr::as_dyn_int(self.arg),
            })
            .intern(),
            __ty: self.arg.__ty.as_same_width_uint(),
        }
    }
}

impl_unary_op_trait! {
    #[generics(Width: GenericConstUsize)]
    fn Not::not(arg: SInt<Width>) -> UInt<Width> {
        NotS { arg }.to_expr()
    }
}

forward_value_to_expr_unary_op_trait! {
    #[generics(Width: GenericConstUsize)]
    #[value(SIntValue<Width>)]
    Not::not
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct NotB {
    pub arg: Expr<Bool>,
}

impl ToExpr for NotB {
    type Type = Bool;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::NotB(NotB { arg: self.arg }).intern(),
            __ty: self.arg.__ty,
        }
    }
}

impl_unary_op_trait! {
    #[generics()]
    fn Not::not(arg: Bool) -> Bool {
        NotB { arg }.to_expr()
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct Neg {
    pub arg: Expr<SInt>,
}

impl ToExpr for Neg {
    type Type = SInt;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::Neg(*self).intern(),
            __ty: SInt::new_dyn(
                Expr::ty(self.arg)
                    .width()
                    .checked_add(1)
                    .expect("width too big"),
            ),
        }
    }
}

impl_unary_op_trait! {
    #[generics(Width: GenericConstUsize)]
    fn StdNeg::neg(arg: SInt<Width>) -> SInt {
        Neg { arg: Expr::as_dyn_int(arg) }.to_expr()
    }
}

forward_value_to_expr_unary_op_trait! {
    #[generics(Width: GenericConstUsize)]
    #[value(SIntValue<Width>)]
    StdNeg::neg
}

macro_rules! impl_binary_op_trait {
    (
        #[generics($($generics:tt)*)]
        fn $Trait:ident::$method:ident($lhs:ident: $Lhs:ty, $rhs:ident: $Rhs:ty) -> $Output:ty {
            $($body:tt)*
        }
    ) => {
        impl<
                Rhs: ToExpr<Type = $Rhs>,
                $($generics)*
            > $Trait<Rhs> for Expr<$Lhs>
        {
            type Output = Expr<$Output>;

            fn $method(self, rhs: Rhs) -> Self::Output {
                let $lhs = self;
                let $rhs = rhs.to_expr();
                $($body)*
            }
        }
    };
}

macro_rules! forward_value_to_expr_binary_op_trait {
    (
        #[generics($($generics:tt)*)]
        #[lhs_value($LhsValue:ty)]
        $Trait:ident::$method:ident
    ) => {
        impl<
                Rhs,
                $($generics)*
            > $Trait<Rhs> for $LhsValue
        where
            Expr<<$LhsValue as ToExpr>::Type>: $Trait<Rhs>,
        {
            type Output = <Expr<<$LhsValue as ToExpr>::Type> as $Trait<Rhs>>::Output;

            fn $method(self, rhs: Rhs) -> Self::Output {
                $Trait::$method(self.to_expr(), rhs)
            }
        }
    };
}

macro_rules! binary_op_bitwise {
    ($name:ident, $ty:ident, bool, $Trait:ident::$method:ident) => {
        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub struct $name {
            pub lhs: Expr<$ty>,
            pub rhs: Expr<$ty>,
        }

        impl ToExpr for $name {
            type Type = $ty;

            fn to_expr(&self) -> Expr<Self::Type> {
                Expr {
                    __enum: ExprEnum::$name(*self).intern(),
                    __ty: $ty,
                }
            }
        }

        impl_binary_op_trait! {
            #[generics()]
            fn $Trait::$method(lhs: $ty, rhs: $ty) -> $ty {
                $name {
                    lhs,
                    rhs,
                }
                .to_expr()
            }
        }
    };
    ($name:ident, $ty:ident, $value:ident, $Trait:ident::$method:ident) => {
        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub struct $name {
            pub lhs: Expr<$ty>,
            pub rhs: Expr<$ty>,
        }

        impl ToExpr for $name {
            type Type = UInt;

            fn to_expr(&self) -> Expr<Self::Type> {
                Expr {
                    __enum: ExprEnum::$name(*self).intern(),
                    __ty: UInt::new_dyn(Expr::ty(self.lhs).width().max(Expr::ty(self.lhs).width())),
                }
            }
        }

        impl_binary_op_trait! {
            #[generics(LhsWidth: GenericConstUsize, RhsWidth: GenericConstUsize)]
            fn $Trait::$method(lhs: $ty<LhsWidth>, rhs: $ty<RhsWidth>) -> UInt {
                $name {
                    lhs: Expr::as_dyn_int(lhs),
                    rhs: Expr::as_dyn_int(rhs),
                }
                .to_expr()
            }
        }

        forward_value_to_expr_binary_op_trait! {
            #[generics(LhsWidth: GenericConstUsize)]
            #[lhs_value($value<LhsWidth>)]
            $Trait::$method
        }
    };
}

binary_op_bitwise!(BitAndU, UInt, UIntValue, BitAnd::bitand);
binary_op_bitwise!(BitAndS, SInt, SIntValue, BitAnd::bitand);
binary_op_bitwise!(BitAndB, Bool, bool, BitAnd::bitand);
binary_op_bitwise!(BitOrU, UInt, UIntValue, BitOr::bitor);
binary_op_bitwise!(BitOrS, SInt, SIntValue, BitOr::bitor);
binary_op_bitwise!(BitOrB, Bool, bool, BitOr::bitor);
binary_op_bitwise!(BitXorU, UInt, UIntValue, BitXor::bitxor);
binary_op_bitwise!(BitXorS, SInt, SIntValue, BitXor::bitxor);
binary_op_bitwise!(BitXorB, Bool, bool, BitXor::bitxor);

macro_rules! binary_op_add_sub {
    ($name:ident, $ty:ident, $value:ident, $Trait:ident::$method:ident) => {
        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub struct $name {
            pub lhs: Expr<$ty>,
            pub rhs: Expr<$ty>,
        }

        impl ToExpr for $name {
            type Type = $ty;

            fn to_expr(&self) -> Expr<Self::Type> {
                Expr {
                    __enum: ExprEnum::$name(*self).intern(),
                    __ty: $ty::new_dyn(
                        Expr::ty(self.lhs)
                            .width()
                            .max(Expr::ty(self.rhs).width())
                            .checked_add(1)
                            .expect("width too big"),
                    ),
                }
            }
        }

        impl_binary_op_trait! {
            #[generics(LhsWidth: GenericConstUsize, RhsWidth: GenericConstUsize)]
            fn $Trait::$method(lhs: $ty<LhsWidth>, rhs: $ty<RhsWidth>) -> $ty {
                $name {
                    lhs: Expr::as_dyn_int(lhs),
                    rhs: Expr::as_dyn_int(rhs),
                }
                .to_expr()
            }
        }

        forward_value_to_expr_binary_op_trait! {
            #[generics(LhsWidth: GenericConstUsize)]
            #[lhs_value($value<LhsWidth>)]
            $Trait::$method
        }
    };
}

binary_op_add_sub!(AddU, UInt, UIntValue, Add::add);
binary_op_add_sub!(AddS, SInt, SIntValue, Add::add);
binary_op_add_sub!(SubU, UInt, UIntValue, Sub::sub);
binary_op_add_sub!(SubS, SInt, SIntValue, Sub::sub);

macro_rules! binary_op_mul {
    ($name:ident, $ty:ident, $value:ident) => {
        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub struct $name {
            pub lhs: Expr<$ty>,
            pub rhs: Expr<$ty>,
        }

        impl ToExpr for $name {
            type Type = $ty;

            fn to_expr(&self) -> Expr<Self::Type> {
                Expr {
                    __enum: ExprEnum::$name(*self).intern(),
                    __ty: $ty::new_dyn(
                        Expr::ty(self.lhs)
                            .width()
                            .checked_add(Expr::ty(self.rhs).width())
                            .expect("width too big"),
                    ),
                }
            }
        }

        impl_binary_op_trait! {
            #[generics(LhsWidth: GenericConstUsize, RhsWidth: GenericConstUsize)]
            fn Mul::mul(lhs: $ty<LhsWidth>, rhs: $ty<RhsWidth>) -> $ty {
                $name {
                    lhs: Expr::as_dyn_int(lhs),
                    rhs: Expr::as_dyn_int(rhs),
                }
                .to_expr()
            }
        }

        forward_value_to_expr_binary_op_trait! {
            #[generics(LhsWidth: GenericConstUsize)]
            #[lhs_value($value<LhsWidth>)]
            Mul::mul
        }
    };
}

binary_op_mul!(MulU, UInt, UIntValue);
binary_op_mul!(MulS, SInt, SIntValue);

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct DivU<LhsWidth: GenericConstUsize = UnknownSize> {
    pub lhs: Expr<UInt<LhsWidth>>,
    pub rhs: Expr<UInt>,
}

impl<LhsWidth: GenericConstUsize> ToExpr for DivU<LhsWidth> {
    type Type = UInt<LhsWidth>;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::DivU(DivU {
                lhs: Expr::as_dyn_int(self.lhs),
                rhs: self.rhs,
            })
            .intern(),
            __ty: Expr::ty(self.lhs),
        }
    }
}

impl_binary_op_trait! {
    #[generics(LhsWidth: GenericConstUsize, RhsWidth: GenericConstUsize)]
    fn Div::div(lhs: UInt<LhsWidth>, rhs: UInt<RhsWidth>) -> UInt<LhsWidth> {
        DivU {
            lhs,
            rhs: Expr::as_dyn_int(rhs),
        }
        .to_expr()
    }
}

forward_value_to_expr_binary_op_trait! {
    #[generics(LhsWidth: GenericConstUsize)]
    #[lhs_value(UIntValue<LhsWidth>)]
    Div::div
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct DivS {
    pub lhs: Expr<SInt>,
    pub rhs: Expr<SInt>,
}

impl ToExpr for DivS {
    type Type = SInt;

    fn to_expr(&self) -> Expr<Self::Type> {
        Expr {
            __enum: ExprEnum::DivS(*self).intern(),
            __ty: SInt::new_dyn(
                Expr::ty(self.lhs)
                    .width()
                    .checked_add(1)
                    .expect("width too big"),
            ),
        }
    }
}

impl_binary_op_trait! {
    #[generics(LhsWidth: GenericConstUsize, RhsWidth: GenericConstUsize)]
    fn Div::div(lhs: SInt<LhsWidth>, rhs: SInt<RhsWidth>) -> SInt {
        DivS {
            lhs: Expr::as_dyn_int(lhs),
            rhs: Expr::as_dyn_int(rhs),
        }
        .to_expr()
    }
}

forward_value_to_expr_binary_op_trait! {
    #[generics(LhsWidth: GenericConstUsize)]
    #[lhs_value(SIntValue<LhsWidth>)]
    Div::div
}

macro_rules! binary_op_rem {
    ($name:ident, $ty:ident, $value:ident) => {
        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub struct $name {
            pub lhs: Expr<$ty>,
            pub rhs: Expr<$ty>,
        }

        impl ToExpr for $name {
            type Type = $ty;

            fn to_expr(&self) -> Expr<Self::Type> {
                Expr {
                    __enum: ExprEnum::$name(*self).intern(),
                    __ty: $ty::new_dyn(Expr::ty(self.lhs).width().min(Expr::ty(self.rhs).width())),
                }
            }
        }

        impl_binary_op_trait! {
            #[generics(LhsWidth: GenericConstUsize, RhsWidth: GenericConstUsize)]
            fn Rem::rem(lhs: $ty<LhsWidth>, rhs: $ty<RhsWidth>) -> $ty {
                $name {
                    lhs: Expr::as_dyn_int(lhs),
                    rhs: Expr::as_dyn_int(rhs),
                }
                .to_expr()
            }
        }

        forward_value_to_expr_binary_op_trait! {
            #[generics(LhsWidth: GenericConstUsize)]
            #[lhs_value($value<LhsWidth>)]
            Rem::rem
        }
    };
}

binary_op_rem!(RemU, UInt, UIntValue);
binary_op_rem!(RemS, SInt, SIntValue);
