// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    intern::{Intern, Interned},
    ty::{CanonicalType, StaticType, Type},
    util::{ConstBool, ConstUsize, GenericConstBool, GenericConstUsize},
};
use bitvec::slice::BitSlice;
use std::{marker::PhantomData, ops::Index, sync::Arc};

pub const UNKNOWN_SIZE: usize = !0;
pub type UnknownSize = ConstUsize<UNKNOWN_SIZE>;

pub trait KnownSize: GenericConstUsize {}

macro_rules! known_widths {
    ([] $($bits:literal)+) => {
        impl KnownSize for ConstUsize<{
            let v = 0;
            $(let v = v * 2 + $bits;)*
            v
        }> {}
    };
    ([2 $($rest:tt)*] $($bits:literal)+) => {
        known_widths!([$($rest)*] $($bits)* 0);
        known_widths!([$($rest)*] $($bits)* 1);
    };
    ([2 $($rest:tt)*]) => {
        known_widths!([$($rest)*] 0);
        known_widths!([$($rest)*] 1);
        impl KnownSize for ConstUsize<{2 $(* $rest)*}> {}
    };
}

known_widths!([2 2 2 2 2 2 2 2 2]);

macro_rules! impl_int {
    ($name:ident, $generic_name:ident, $value:ident, $SIGNED:literal) => {
        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub struct $name<Width: GenericConstUsize = UnknownSize> {
            width: usize,
            _phantom: PhantomData<Width>,
        }

        #[allow(non_upper_case_globals)]
        pub const $name: $generic_name = $generic_name;

        impl<Width: GenericConstUsize> $name<Width> {
            pub const fn width(self) -> usize {
                if Width::VALUE != UNKNOWN_SIZE {
                    debug_assert!(self.width == Width::VALUE);
                    Width::VALUE
                } else {
                    self.width
                }
            }
        }

        impl<Width: GenericConstUsize> sealed::IntTypeSealed for $name<Width> {}

        impl<Width: GenericConstUsize> IntType for $name<Width> {
            type Dyn = $name;
            type Width = Width;
            type Signed = ConstBool<$SIGNED>;
            fn width(self) -> usize {
                $name::width(self)
            }
            fn new_dyn(width: usize) -> Self::Dyn {
                $name::new_dyn(width)
            }
            fn new_static() -> Self
            where
                Self::Width: KnownSize,
            {
                $name::new_static()
            }
            fn as_same_width_sint(self) -> SInt<Self::Width> {
                SInt {
                    width: self.width(),
                    _phantom: PhantomData,
                }
            }
            fn as_same_width_uint(self) -> UInt<Self::Width> {
                UInt {
                    width: self.width(),
                    _phantom: PhantomData,
                }
            }
        }

        impl $name {
            pub const fn new_dyn(width: usize) -> Self {
                assert!(width != UNKNOWN_SIZE);
                Self {
                    width,
                    _phantom: PhantomData,
                }
            }
        }

        impl<Width: KnownSize> $name<Width> {
            pub const fn new_static() -> Self {
                Self {
                    width: Width::VALUE,
                    _phantom: PhantomData,
                }
            }
        }

        impl<Width: GenericConstUsize> Type for $name<Width> {
            fn canonical(&self) -> CanonicalType {
                CanonicalType::$name($name::new_dyn(self.width()))
            }
        }

        impl<Width: KnownSize> StaticType for $name<Width> {
            fn static_type() -> Self {
                Self::new_static()
            }
        }

        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug, Default)]
        pub struct $generic_name;

        impl Index<usize> for $generic_name {
            type Output = $name;

            fn index(&self, width: usize) -> &Self::Output {
                Interned::<_>::into_inner(Intern::intern_sized($name::new_dyn(width)))
            }
        }

        #[derive(Clone, PartialEq, Eq, Hash, Debug)]
        pub struct $value<Width: GenericConstUsize = UnknownSize> {
            bits: Arc<BitSlice>,
            _phantom: PhantomData<Width>,
        }

        impl<Width: GenericConstUsize> $value<Width> {
            pub fn width(&self) -> usize {
                if Width::VALUE != UNKNOWN_SIZE {
                    debug_assert!(self.bits.len() == Width::VALUE);
                    Width::VALUE
                } else {
                    self.bits.len()
                }
            }
            pub fn ty(&self) -> $name<Width> {
                $name {
                    width: self.width(),
                    _phantom: PhantomData,
                }
            }
            pub fn as_dyn_int(self) -> $value {
                $value {
                    bits: self.bits,
                    _phantom: PhantomData,
                }
            }
        }

        impl<Width: KnownSize> $value<Width> {
            pub fn new_static(bits: Arc<BitSlice>) -> Self {
                assert!(bits.len() == Width::VALUE);
                Self {
                    bits,
                    _phantom: PhantomData,
                }
            }
        }

        impl $value {
            pub fn new_dyn(bits: Arc<BitSlice>) -> Self {
                Self {
                    bits,
                    _phantom: PhantomData,
                }
            }
        }
    };
}

impl_int!(UInt, UIntWithoutGenerics, UIntValue, false);
impl_int!(SInt, SIntWithoutGenerics, SIntValue, true);

mod sealed {
    pub trait IntTypeSealed {}
}

pub trait IntType: Type + sealed::IntTypeSealed {
    type Dyn: IntType<Dyn = Self::Dyn, Signed = Self::Signed, Width = UnknownSize>;
    type Signed: GenericConstBool;
    type Width: GenericConstUsize;
    fn width(self) -> usize;
    fn as_dyn_int(self) -> Self::Dyn {
        Self::new_dyn(self.width())
    }
    fn new_dyn(width: usize) -> Self::Dyn;
    fn new_static() -> Self
    where
        Self::Width: KnownSize;
    fn as_same_width_sint(self) -> SInt<Self::Width>;
    fn as_same_width_uint(self) -> UInt<Self::Width>;
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct Bool;

impl Type for Bool {
    fn canonical(&self) -> CanonicalType {
        CanonicalType::Bool(Bool)
    }
}

impl StaticType for Bool {
    fn static_type() -> Self {
        Bool
    }
}
